/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author Gareth Johnson
 */

/**
 * @module env/Promises
 * @private
 */
define('bajaPromises', [ 'Promise' ], function (Promise) {
  'use strict';

  /**
   * Promise wrapper that exports promise functions needed by BajaScript.
   * You can swap in your own implementation if you need to use a different
   * Promise library. It must implement this exact API.
   * 
   * Specify it in your RequireJS config as a RequireJS ID at
   * `baja.config.promises`.
   * 
   * @exports bajaScript/env/Promises
   */
  return {
    /**
     * Return a promise that has been fulfilled with the given value.
     * @param {*} obj
     * @returns {Promise}
     */
    resolve: function (obj) {
      return Promise.resolve(obj);
    },

    /**
     * Return a promise that has been rejected with the given error.
     * @param {*} err
     * @returns {Promise}
     */
    reject: function (err) {
      return Promise.reject(err);
    },

    /**
     * Return a Deferred object. It should have three methods:
     * `resolve` that resolves the promise, `reject` that rejects the promise,
     * and `promise` that returns the Promise itself.
     * @returns {Object}
     */
    deferred: function () {
      var resolve,
          reject,
          promise = new Promise(function (res, rej) {
            resolve = res; reject = rej;
          });
      
      return {
        resolve:resolve,
        reject: reject,
        promise: function () { return promise; }
      };
    },

    /**
     * Return a promise that is resolved when all the elements of the array
     * (values or Promises) are resolved.
     * @param {Array.<*|Promise>} array
     * @returns {Promise}
     */
    all: function (array) {
      return Promise.all(array);
    },

    /**
     * Return true if the given promise has been resolved.
     * 
     * @param {Promise} prom
     * @returns {Boolean}
     */
    isResolved: function (prom) {
      return prom.isFulfilled();
    },

    /**
     * Return true if the given promise has been rejected.
     * 
     * @param {Promise} prom
     * @returns {Boolean}
     */
    isRejected: function (prom) {
      return prom.isRejected();
    },

    /**
     * Return true if the given promise is still pending.
     * 
     * @param {Promise} prom
     * @returns {Boolean}
     */
    isPending: function (prom) {
      return prom.isPending();
    }
  };
});